/*
 * Copyright 2004 The Unichrome Project  [unichrome.sf.net]
 * Copyright 1998-2003 VIA Technologies, Inc. All Rights Reserved.
 * Copyright 2001-2003 S3 Graphics, Inc. All Rights Reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sub license,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice (including the
 * next paragraph) shall be included in all copies or substantial portions
 * of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL
 * VIA, S3 GRAPHICS, AND/OR ITS SUPPLIERS BE LIABLE FOR ANY CLAIM, DAMAGES OR
 * OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 */
 
#include "xf86.h"
#include "via.h"
#include "via_driver.h"
#include "via_overlay.h"
#include "via_id.h"
#include <math.h>

/* Verify that using V1 bit definitions on V3
 * is not broken in OverlayGetV1V3Format()
 */

#if V1_COLORSPACE_SIGN != V3_COLORSPACE_SIGN
#error "V1_COLORSPACE_SIGN != V3_COLORSPACE_SIGN"
#endif
#if V1_YUV422 != V3_YUV422
#error "V1_YUV422 != V3_YUV422"
#endif
#if V1_SWAP_HW_HQV != V3_SWAP_HW_HQV
#error "V1_SWAP_HW_HQV != V3_SWAP_HW_HQV"
#endif
#if V1_RGB15 != V3_RGB15
#error "V1_RGB15 != V3_RGB15"
#endif
#if V1_RGB16 != V3_RGB16
#error "V1_RGB16 != V3_RGB16"
#endif
#if V1_RGB32 != V3_RGB32
#error "V1_RGB32 != V3_RGB32"
#endif

BOOL viaOverlayGetV1V3Format(VIAPtr pVia, int vport, /* 1 or 3, as in V1 or V3 */
                             unsigned long videoFlag, LPDDPIXELFORMAT pPF,
                             unsigned long * pVidCtl, unsigned long * pHQVCtl)
{
    BOOL is_ok = TRUE;

    if (pPF->dwFlags & DDPF_FOURCC)
    {
        *pVidCtl |= V1_COLORSPACE_SIGN;
        switch (pPF->dwFourCC)
        {
        case FOURCC_YV12:
        case FOURCC_VIA:
            if (videoFlag & VIDEO_HQV_INUSE) {
                *pVidCtl |= V1_YUV422 | V1_SWAP_HW_HQV;
                *pHQVCtl |= HQV_SRC_SW | HQV_YUV420 | HQV_ENABLE | HQV_SW_FLIP;
            }
            else {
                if (vport == 1) {
                    *pVidCtl |= V1_YCbCr420;
                }
                else {
                    DBG_DD(ErrorF("viaOverlayGetV1V3Format: V3 does not support planar YUV.\n"));
                    is_ok = FALSE;
                }
            }
            break;

        case FOURCC_YUY2:
            if (videoFlag & VIDEO_HQV_INUSE) {
                *pVidCtl |= V1_YUV422 | V1_SWAP_HW_HQV;
                *pHQVCtl |= HQV_SRC_SW | HQV_YUV422 | HQV_ENABLE | HQV_SW_FLIP;
            }
            else {
                *pVidCtl |= V1_YUV422;
            }
            break;

        default:
            DBG_DD(ErrorF("viaOverlayGetV1V3Format: Invalid FOURCC format (0x%lx).\n",
                pPF->dwFourCC));
            is_ok = FALSE;

            *pVidCtl |= V1_YUV422;
            break;
        }
    }
    else if (pPF->dwFlags & DDPF_RGB)
    {
        switch (pPF->dwRGBBitCount) 
        {
        case 16:
            if (videoFlag & VIDEO_HQV_INUSE) {
                *pVidCtl |= V1_RGB16 | V1_SWAP_HW_HQV;
                *pHQVCtl |= HQV_SRC_SW | HQV_ENABLE | HQV_SW_FLIP;
                *pHQVCtl |= (pPF->dwGBitMask==0x07E0 ? HQV_RGB16 : HQV_RGB15);
            }
            else {
                *pVidCtl |= (pPF->dwGBitMask==0x07E0 ? V1_RGB16 : V1_RGB15);
            }
            break;

        case 32:
            if (videoFlag & VIDEO_HQV_INUSE) {
                *pVidCtl |= V1_RGB32 | V1_SWAP_HW_HQV;
                *pHQVCtl |= HQV_SRC_SW | HQV_RGB32 | HQV_ENABLE | HQV_SW_FLIP;
            }
            else {
                *pVidCtl |= V1_RGB32;
            }
            break;

        default:
            DBG_DD(ErrorF("viaOverlayGetV1V3Format: invalid RGB format (%ld bits).\n",
                pPF->dwRGBBitCount));
            is_ok = FALSE;
            break;
        }
    }

    return is_ok;
}

unsigned long viaOverlayGetSrcStartAddress(VIAPtr pVia, unsigned long videoFlag,
                                           RECTL rSrc, RECTL rDst, unsigned long srcPitch,
                                           LPDDPIXELFORMAT pPF, unsigned long * pHQVoffset)
{
    unsigned long srcWidth  = (unsigned long)(rSrc.right - rSrc.left);
    unsigned long dstWidth  = (unsigned long)(rDst.right - rDst.left);
    unsigned long srcHeight = (unsigned long)(rSrc.bottom - rSrc.top);
    unsigned long dstHeight = (unsigned long)(rDst.bottom - rDst.top);

    unsigned long offset = 0;
    unsigned long srcTopOffset = 0;
    unsigned long srcLeftOffset = 0;

    if ( (rSrc.left != 0) || (rSrc.top != 0) )
    {
        if (pPF->dwFlags & DDPF_FOURCC)
        {
            switch (pPF->dwFourCC)
            {
            case FOURCC_YUY2:
            case FOURCC_UYVY:

                if (videoFlag & VIDEO_HQV_INUSE) {
                    offset = (((rSrc.top & ~3) * srcPitch) +
                        ((rSrc.left << 1) & ~31));

                    if (srcHeight > dstHeight) {
                        srcTopOffset = ((rSrc.top & ~3) * dstHeight / srcHeight) * srcPitch;
                    }
                    else {
                        srcTopOffset = (rSrc.top & ~3) * srcPitch;
                    }

                    if (srcWidth > dstWidth) {
                        srcLeftOffset = ((rSrc.left << 1) & ~31) * dstWidth / srcWidth;
                    }
                    else {
                        srcLeftOffset = (rSrc.left << 1) & ~31;
                    }
                    *pHQVoffset = srcTopOffset + srcLeftOffset;
                }
                else {
                    offset = ((rSrc.top * srcPitch) + ((rSrc.left << 1) & ~15));
                }
                break;

            case FOURCC_YV12:
            case FOURCC_VIA:

                if (videoFlag & VIDEO_HQV_INUSE) {
                    offset = (((rSrc.top & ~3) * (srcPitch << 1)) +
                        ((rSrc.left << 1) & ~31));
                }
                else {
                    offset = ((((rSrc.top & ~3) * srcPitch) + rSrc.left) & ~31);
                    if (rSrc.top > 0) {
                        pVia->swov.overlayRecordV1.dwUVoffset =
                            (((((rSrc.top & ~3)>>1) * srcPitch) + rSrc.left) & ~31) >> 1;
                    }
                    else {
                        pVia->swov.overlayRecordV1.dwUVoffset = offset >> 1;
                    }
                }
                break;

            default:
                DBG_DD(ErrorF("viaGetSrcStartAddress: Invalid FOURCC format (0x%lx).\n",
                    pPF->dwFourCC));
                break;
            }
        }
        else if (pPF->dwFlags & DDPF_RGB)
        {
            if (videoFlag & VIDEO_HQV_INUSE)
            {
                offset = (((rSrc.top & ~3) * (srcPitch << 1)) + ((rSrc.left << 1) & ~31));

                if (srcHeight > dstHeight) {
                    srcTopOffset = ((rSrc.top & ~3) * dstHeight / srcHeight) * srcPitch;
                }
                else {
                    srcTopOffset = (rSrc.top & ~3) * srcPitch;
                }

                if (srcWidth > dstWidth) {
                    srcLeftOffset = ((rSrc.left << 1) & ~31) * dstWidth / srcWidth;
                }
                else {
                    srcLeftOffset = (rSrc.left << 1) & ~31;
                }
                *pHQVoffset = srcTopOffset + srcLeftOffset;

            }
            else {
                offset = (rSrc.top * srcPitch) + ((rSrc.left * pPF->dwRGBBitCount) >> 3);
            }
        }
    }
    else {
        pVia->swov.overlayRecordV1.dwUVoffset = offset = 0;
    }

    return offset;
}

YCBCRREC viaOverlayGetYCbCrStartAddress(unsigned long videoFlag,
                                        unsigned long startAddr,
                                        unsigned long offset,
                                        unsigned long UVoffset,
                                        unsigned long srcPitch  /*lpGbl->lPitch*/,
                                        unsigned long srcHeight /*lpGbl->wHeight*/)
{
    YCBCRREC YCbCr;

    if (videoFlag & VIDEO_HQV_INUSE) {
        YCbCr.dwY  = startAddr;
        YCbCr.dwCB = startAddr + srcPitch * srcHeight;
        YCbCr.dwCR = startAddr + srcPitch * srcHeight
            + srcPitch * (srcHeight >> 2);
    }
    else {
        YCbCr.dwY  = startAddr + offset;
        YCbCr.dwCB = startAddr + srcPitch * srcHeight + UVoffset;
        YCbCr.dwCR = startAddr + srcPitch * srcHeight + UVoffset
            + srcPitch * (srcHeight >> 2);
    }
    return YCbCr;
}

unsigned long viaOverlayHQVCalcZoomWidth(VIAPtr pVia,
                                         unsigned long videoFlag,
                                         unsigned long srcWidth,
                                         unsigned long dstWidth,
                                         unsigned long * pZoomCtl,
                                         unsigned long * pMiniCtl,
                                         unsigned long * pHQVfilterCtl,
                                         unsigned long * pHQVminiCtl,
                                         unsigned long * pHQVzoomflag)
{
    unsigned long tmp, sw1, d, falign, mdiv;
    Bool zoom_ok = TRUE;

    CARD32 HQVfilter[5] = {
        HQV_H_FILTER_DEFAULT, HQV_H_TAP4_121, HQV_H_TAP4_121,
        HQV_H_TAP8_12221, HQV_H_TAP8_12221
    };
    /* CARD HQVmini[5] = { 0, 0xc00, 0xa00, 0x900, 0x8800 }; */

    falign = 0;
    mdiv = 1;

    if (srcWidth == dstWidth) { /* No zoom */
        *pHQVfilterCtl |= HQV_H_FILTER_DEFAULT;
    }
    else if (srcWidth < dstWidth) { /* Zoom in */

        tmp = srcWidth * 0x800 / dstWidth;
        *pZoomCtl = ((tmp & 0x7ff) << 16) | V1_X_ZOOM_ENABLE;
        *pMiniCtl |= V1_X_INTERPOLY;
        zoom_ok = !(tmp > 0x7ff);

        *pHQVzoomflag = 1;
        *pHQVfilterCtl |= HQV_H_FILTER_DEFAULT;

    } else { /* srcWidth > dstWidth - Zoom out */

        /*HQV rounding patch
        //tmp = dstWidth*0x0800 / srcWidth;*/
        tmp = dstWidth * 0x800 * 0x400 / srcWidth;
        tmp = tmp / 0x400 + ((tmp & 0x3ff) ? 1 : 0);

        *pHQVminiCtl = (tmp & 0x7ff) | HQV_H_MINIFY_ENABLE;

        /* Scale down the picture by a factor mdiv = (1 << d) = {2, 4, 8 or 16} */

        sw1 = srcWidth;
        for (d = 1; d < 5; d++) {
            sw1 >>= 1;
            if (sw1 <= dstWidth) break;
        }
        if (d == 5) { /* Too small. */
            d = 4;
            zoom_ok = FALSE;
        }
        mdiv = 1 << d;                  /* <= {2,4,8,16} */
        falign = ((mdiv<<1)-1) & 0xf;   /* <= {3,7,15,15} */
        *pMiniCtl |= V1_X_INTERPOLY;
        *pMiniCtl |= ((d<<1)-1) << 24;  /* <= {1,3,5,7} << 24 */

        *pHQVfilterCtl |= HQVfilter[d];
        /* *pHQVminiCtl = HQVmini[d]; */
        *pHQVminiCtl |= HQV_HDEBLOCK_FILTER;

        /* Scale to arbitrary size, on top of previous scaling by (1 << d). */

        if (sw1 < dstWidth) {
            /* CLE bug
            *pZoomCtl = sw1 * 0x0800 / dstWidth;*/
            *pZoomCtl = (sw1 - 2) * 0x0800 / dstWidth;
            *pZoomCtl = ((*pZoomCtl & 0x7ff) << 16) | V1_X_ZOOM_ENABLE;
        }
    }

    if (videoFlag & VIDEO_1_INUSE) {
        pVia->swov.overlayRecordV1.dwFetchAlignment = falign;
        pVia->swov.overlayRecordV1.dwminifyH = mdiv;
    }
    else {
        pVia->swov.overlayRecordV3.dwFetchAlignment = falign;
        pVia->swov.overlayRecordV3.dwminifyH = mdiv;
    }

    return zoom_ok;
}

unsigned long viaOverlayHQVCalcZoomHeight(VIAPtr pVia,
                                          unsigned long srcHeight,
                                          unsigned long dstHeight,
                                          unsigned long * pZoomCtl,
                                          unsigned long * pMiniCtl,
                                          unsigned long * pHQVfilterCtl,
                                          unsigned long * pHQVminiCtl,
                                          unsigned long * pHQVzoomflag)
{
    unsigned long tmp, sh1, d;
    Bool zoom_ok = TRUE;

    CARD32 HQVfilter[5] = {
        HQV_V_TAP4_121, HQV_V_TAP4_121, HQV_V_TAP4_121,
        HQV_V_TAP8_12221, HQV_V_TAP8_12221
    };
    /* CARD32 HQVmini[5] = { 0, 0x0c000000, 0x0a000000, 0x09000000, 0x08800000 }; */

    /*if (pVia->pBIOSInfo->scaleY)
    {
        dstHeight = dstHeight + 1;
    }*/

    if (srcHeight == dstHeight) { /* No zoom */
        *pHQVfilterCtl |= HQV_V_TAP4_121;
    }
    else if (srcHeight < dstHeight) { /* Zoom in */

        tmp = srcHeight * 0x0400 / dstHeight;
        *pZoomCtl |= ((tmp & 0x3ff) | V1_Y_ZOOM_ENABLE);
        *pMiniCtl |= (V1_Y_INTERPOLY | V1_YCBCR_INTERPOLY);

        *pHQVzoomflag = 1;
        *pHQVfilterCtl |= HQV_V_TAP4_121;
    } 
    else { /* srcHeight > dstHeight - Zoom out */

        /*HQV rounding patch
        //tmp = dstHeight*0x0800 / srcHeight;*/
        tmp = dstHeight * 0x0800 * 0x400 / srcHeight;
        tmp = tmp / 0x400 + ((tmp & 0x3ff) ? 1 : 0);
        *pHQVminiCtl |= ((tmp& 0x7ff) << 16) | HQV_V_MINIFY_ENABLE;

        /* Scale down the picture by a factor (1 << d) = {2, 4, 8 or 16} */

        sh1 = srcHeight;
        for (d = 1; d < 5; d++) {
            sh1 >>= 1;
            if (sh1 <= dstHeight) break;
        }
        if (d == 5) { /* Too small. */
            d = 4;
            zoom_ok = FALSE;
        }

        *pMiniCtl |= ((d<<1)-1) << 16;  /* <= {1,3,5,7} << 16 */

        *pHQVfilterCtl |= HQVfilter[d];
        /* *pHQVminiCtl |= HQVmini[d]; */
        *pHQVminiCtl |= HQV_VDEBLOCK_FILTER;

        /* Scale to arbitrary size, on top of previous scaling by (1 << d). */

        if (sh1 < dstHeight) {
            tmp = sh1 * 0x0400 / dstHeight;
            *pZoomCtl |= ((tmp & 0x3ff) | V1_Y_ZOOM_ENABLE);
            *pMiniCtl |= V1_Y_INTERPOLY | V1_YCBCR_INTERPOLY;
        }
    }

    return zoom_ok;
}

unsigned long viaOverlayGetFetch(unsigned long videoFlag, LPDDPIXELFORMAT pPF,
                                 unsigned long srcWidth, unsigned long dstWidth,
                                 unsigned long oriSrcWidth, unsigned long * pHQVsrcFetch)
{
    unsigned long fetch = 0;
    int n = 2; /* 2^n bytes per pixel. */

    if (pPF->dwFlags & DDPF_FOURCC)
    {
        switch (pPF->dwFourCC)
        {
        case FOURCC_YV12:
        case FOURCC_VIA:
            n = 0; /* 2^n = 1 byte per pixel (Y channel in planar YUV) */
            break;
        case FOURCC_UYVY:
        case FOURCC_YUY2:
            n = 1; /* 2^n = 2 bytes per pixel (packed YUV) */
            break;
        default:
            DBG_DD(ErrorF("viaOverlayGetFetch: Invalid FOURCC format (0x%lx).\n",
                pPF->dwFourCC));
            break;
        }
    }
    else if (pPF->dwFlags & DDPF_RGB)
    {
        switch (pPF->dwRGBBitCount)
        {
        case 16:
            n = 1; /* 2^n = 2 bytes per pixel (RGB16) */
            break;
        case 32:
            n = 2; /* 2^n = 4 bytes per pixel (RGB32) */
            break;
        default:
            DBG_DD(ErrorF("viaOverlayGetFetch: invalid RGB format (%ld bits).\n",
                pPF->dwRGBBitCount));
            break;
        }
    }

    if (videoFlag & VIDEO_HQV_INUSE) {
        *pHQVsrcFetch = oriSrcWidth << n;
        if (n == 0) {
            /* Assume n == 0 <=> Planar YUV.
             * The V1/V3 pixelformat is always packed YUV when we use HQV,
             * so we switch from 8-bit to 16-bit pixels here.
             */
            n = 1;
        }
        if (dstWidth >= srcWidth)
            fetch = (ALIGN_TO(srcWidth << n, 16) >> 4) + 1;
        else
            fetch = (ALIGN_TO(dstWidth << n, 16) >> 4) + 1;
    }
    else {
        if (n == 0)
            fetch = (ALIGN_TO(srcWidth, 32) >> 4);
        else
            fetch = (ALIGN_TO(srcWidth << n, 16) >> 4) + 1;
    }

    /* Fix planar mode problem. */
    if (fetch < 4) fetch = 4;

    return fetch;
}

void viaOverlayGetDisplayCount(VIAPtr pVia, unsigned long videoFlag, LPDDPIXELFORMAT pPF,
                               unsigned long srcWidth, unsigned long * pDisplayCountW)
{
    /* Removed VIA's large pixelformat switch/case.
     * All formats (YV12, UYVY, YUY2, VIA, RGB16 and RGB32)
     * seem to use the same count. /A
     */

    if (videoFlag & VIDEO_HQV_INUSE) {
        *pDisplayCountW = srcWidth - 1;
    }
    else {
        *pDisplayCountW = srcWidth - pVia->swov.overlayRecordV1.dwminifyH;
    }
}

/*
 * This function uses quadratic mapping to adjust the midpoint of the scaling.
 */

static float rangeEqualize(float inLow,float inHigh,float outLow,float outHigh,float outMid,
                           float inValue)
{
    float
        inRange = inHigh - inLow,
        outRange = outHigh - outLow,
        normIn = ((inValue - inLow) / inRange)*2.-1.,
        delta = outMid - outRange*0.5 - outLow;
    return (inValue - inLow) * outRange / inRange + outLow + (1. - normIn*normIn)*delta;
}

static unsigned vPackFloat(float val, float hiLimit, float loLimit, float mult, int shift, 
                           Bool doSign)
{
    unsigned packed,mask,sign;
    val = (val > hiLimit) ? hiLimit : val;
    val = (val < loLimit) ? loLimit : val;
    sign = (val < 0) ? 1:0;
    val = (sign) ? -val : val;
    packed = ((unsigned)(val*mult + 1.)) >> 1;
    mask = (1 << shift) - 1;
    return (((packed >= mask) ? mask : packed) | ((doSign) ? (sign << shift) : 0));

}


typedef float colorCoeff[5];
static colorCoeff colorCTable[] = {{1.1875,1.625,0.875,0.375,2.0},
                                   {1.164,1.596,0.54,0.45,2.2}};

/*
 * This function is a partial rewrite of the overlay.c file of the original VIA drivers,
 * which was extremely nasty and difficult to follow. Coefficient for new chipset models should
 * be added in the table above and, if needed, implemented in the model switch below.
 */

void viaCalculateVideoColor(VIAPtr pVia, int hue, int saturation, int brightness, 
                            int contrast,Bool reset,CARD32 *col1,CARD32 *col2)
{
    float fA,fB1,fC1,fD,fB2,fC2,fB3,fC3;
    float fPI,fContrast,fSaturation,fHue,fBrightness;
    const float *mCoeff; 
    unsigned long dwA,dwB1,dwC1,dwD,dwB2,dwC2,dwB3,dwC3,dwS;
    unsigned long dwD_Int,dwD_Dec;
    int intD;
    int model;
    fPI = (float)(M_PI/180.);    

    if ( reset ) {
        saturation = 10000;
        brightness = 5000;
        contrast = 10000;
    }

    switch ( pVia->ChipId )
    {
    case PCI_CHIP_VT3205:
        model = 0;
        break;
    case PCI_CHIP_CLE3122:
        model = (CLE266_REV_IS_CX(pVia->ChipRev) ? 0 : 1); 
        break;
    default:
        ErrorF("Unknown Chip ID\n");
        model = 0;
    }

    switch( model )
    {
    case 0:
        fBrightness = rangeEqualize(0.,10000.,-128.,128.,-16.,(float) brightness);
        fContrast = rangeEqualize(0.,20000.,0.,1.6645,1.0,(float) contrast);
        fSaturation = rangeEqualize(0.,20000,0.,2.,1.,(float) saturation);
        break;
    default:
        fBrightness = rangeEqualize(0.,10000.,-128.,128.,-12.,(float) brightness);
        fContrast = rangeEqualize(0.,20000.,0.,1.6645,1.1,(float) contrast);
        fSaturation = rangeEqualize(0.,20000,0.,2.,1.15,(float) saturation);
        break;
    }
    fHue = (float)hue;

    mCoeff = colorCTable[model];

    fA  = (float)(mCoeff[0]*fContrast);
    fB1 = (float)(-mCoeff[1]*fContrast*fSaturation*sin(fHue*fPI));
    fC1 = (float)(mCoeff[1]*fContrast*fSaturation*cos(fHue*fPI));
    fD  = (float)(mCoeff[0]*(fBrightness));
    fB2 = (float)((mCoeff[2]*sin(fHue*fPI)-
        mCoeff[3]*cos(fHue*fPI))*fContrast*fSaturation);
    fC2 = (float)(-(mCoeff[2]*cos(fHue*fPI)+
        mCoeff[3]*sin(fHue*fPI))*fContrast*fSaturation);
    fB3 = (float)(mCoeff[4]*fContrast*fSaturation*cos(fHue*fPI));
    fC3 = (float)(mCoeff[4]*fContrast*fSaturation*sin(fHue*fPI));

    switch(model)
    {
    case 0:
        dwA = vPackFloat(fA,1.9375,0.,32.,5,0);
        dwB1 = vPackFloat(fB1,2.125,-2.125,16.,5,1);
        dwC1 = vPackFloat(fC1,2.125,-2.125,16.,5,1);

        if (fD>=0) {
            intD=(int)fD;
            if (intD>127)
                intD = 127;
            dwD_Int = ((unsigned long)intD)&0xff;
            dwD = ((unsigned long)(fD*16+1))>>1;
            dwD_Dec= dwD&0x7;
        } else {
            intD=(int)fD;
            if (intD< -128)
                intD = -128;
            intD = intD+256; 
            dwD_Int = ((unsigned long)intD)&0xff;
            fD = -fD;
            dwD = ((unsigned long)(fD*16+1))>>1;
            dwD_Dec= dwD&0x7;
        }

        dwB2 = vPackFloat(fB2,1.875,-1.875,16,4,1);
        dwC2 = vPackFloat(fC2,1.875,-1.875,16,4,1);
        dwB3 = vPackFloat(fB3,3.875,-3.875,16,5,1);
        dwC3 = vPackFloat(fC3,3.875,-3.875,16,5,1);
        *col1 = (dwA<<24)|(dwB1<<16)|(dwC1<<8)|dwD_Int;
        *col2 = (dwD_Dec<<29|dwB2<<24)|(dwC2<<16)|(dwB3<<8)|(dwC3);
        break;

    default:
        dwA = vPackFloat(fA,1.9375,-0.,32,5,0);
        dwB1 = vPackFloat(fB1,0.75,-0.75,8.,2,1);
        dwC1 = vPackFloat(fC1,2.875,1.,16.,5,0);

        if (fD>=127)
            fD=127;

        if (fD<=-128)
            fD=-128;

        if (fD>=0) {
            dwS = 0;
        }
        else {
            dwS = 1;
            fD = fD+128;
        }

        dwD = ((unsigned long)(fD*2+1))>>1;
        if (dwD>=0x7f) {
            dwD = 0x7f|(dwS<<7);
        } else {
            dwD = (dwD&0x7f)|(dwS<<7);
        }

        dwB2 = vPackFloat(fB2,0.,-0.875,16.,3,0); 
        dwC2 = vPackFloat(fC2,0.,-1.875,16.,4,0); 
        dwB3 = vPackFloat(fB3,3.75,0.,8.,4,0);
        dwC3 = vPackFloat(fC3,1.25,-1.25,8.,3,1);
        *col1 = (dwA<<24)|(dwB1<<18)|(dwC1<<9)|dwD;
        *col2 = (dwB2<<25)|(dwC2<<17)|(dwB3<<10)|(dwC3<<2);
        break;
    }
}
